#!/usr/bin/perl -w

package klinstall;

# Constants. We don't use 'use constant' - it is perl 5.6 specific.
my $INI_ERROR_TEXT = "";
my $CFG_AUTOANSWERS_FILE = $ENV{"KLAUTOANSWERS"} ? $ENV{"KLAUTOANSWERS"} : "autoanswers.conf";
my $CFG_APPSETUP_FILE = "/var/opt/kaspersky/applications.setup";
my %AUTOANSWERS;

my %PKGMGRS = ("rpm" => "rpm", "deb" => "deb", "pkg" => "pkg", "opkg" => "pkg");
my $IS_FREEBSD = ($^O =~ /^freebsd$/i) ? 1 : 0;

my $RULES_SPLIT_REPLACE_STR = "_REPLACE_"; # used for converter (rules file)
my $RULES_SPLIT_FROM_STR    = "_FROM_"; # used for converter (rules file)
my $RULES_SPLIT_TO_STR      = "_TO_"; # used for converter (rules file)
my $RULES_MULT_STR    = "MUL"; # used for converter (rules file)
my $RULES_MULTLINE_STR    = "LINE"; # used for converter (rules file)

################### module initialization #############################
$| = 1;


# read autoanswers file if present
if ( open FILE, "<$CFG_AUTOANSWERS_FILE" )
{
	foreach (<FILE>)
	{
		s/[\n\r]//g;
		next if !/^(\w+)\s*=\s*(.*)/;
		$AUTOANSWERS{$1} = $2;
	}

	close FILE;
}

# every module should return true to be loaded successfully
1;

################## external subs #########################################

sub app_info_filename
{
    return $CFG_APPSETUP_FILE;
}

sub autoanswer
{
    return $AUTOANSWERS{shift};
}

# Standard formatted output
sub output
{
	my $text = translate_string(@_) . "\n";
    my $right_margin = 65;
	my $margin = $right_margin - 1;
    my ($para, $accum, $outstr);

    foreach $para (split ("\n", $text))
	{
		$accum = '';
		$para =~ s/\s+/ /g;
		$para .= ' ';

		while ($para =~ s/^([^\n]{0,$margin})\s//)
		{
			$accum .= $1 . "\n";
		}

		$accum .= $para if length ($para) > $margin;
		$outstr .= $accum;
    }

    print $outstr if defined $outstr;
	return 1;
}


# Debug output
sub debug
{
	output "DEBUG: " . translate_string(@_) . "\n";
}

# Terminate installation process
sub fatal
{
	output "\nFatal error: " . translate_string(@_) . "\n";
	exit 2;
}

# Generate a warning
sub warning
{
	output "Warning: " . translate_string(@_) . "\n";
}

# Replace %1, %2 etc. by args.
# Args can contain %1 too, but they should not be replaced
sub translate_string
{
	my ($str, @args) = @_;

	# First, convert color sequences
	$str =~ s/<B>/\x1B\x5B1m/g;
	$str =~ s/<\/B>/\x1B\x5B22m/g;

	for ( my $i = 0; $i < scalar @args; $i++ )
	{
		warn ("Empty argument $i in string '$str'\n") if !defined $args[$i];
		my $pattern = "%" . ($i + 1);
		$str =~ s/$pattern/$args[$i]/g;
	}

	return $str;
}

# Return information about Linux distribution
sub get_linux_info
{
	my %ld;

	$ld{"NAME"} = "RedHat";
	$ld{"VERSION"} = "7.2";

	return %ld;
}


# Return application info for current ID
sub read_app_info
{
	my $id = shift;
	my %appinfo;

	# read settings data
	open F, "<$CFG_APPSETUP_FILE" or return;
	flock F, 2; # LOCK_EX
	my $content = join ("", <F>);
	flock F, 8; # LOCK_UN
	close F;
	$content =~ s/\r//g;

	return unless $content =~ /\[$id\]\n(.*?)\n\n/s;
	$content = $1;

	foreach (split (/\n/, $content))
	{
		next if !/^(.*?)=(.*)$/;

		if ( !defined $appinfo{$1} )
		{
			$appinfo{$1} = $2;
			next;
		}

		# Create an array ref
		if ( ref $appinfo{$1} ne "ARRAY" )
		{
			my @arr = ($appinfo{$1}, $2);
			$appinfo{$1} = \@arr;
		}
		else
		{
			push @{$appinfo{$1}}, $2;
		}
	}

	close F;
	return %appinfo;
}


# Stores application info for specified ID (or adds the existing info)
sub write_app_info
{
	my ($appid, %appinfo) = @_;
	my $content = "";

	# store all the data in text string
	my %olddata = read_app_info ($appid);
	%appinfo = (%olddata, %appinfo) if %olddata;

	my $newdata = "[$appid]\n";
	foreach my $key (sort keys %appinfo)
	{
		my $value = defined $appinfo{$key} ? $appinfo{$key} : "";

		if ( ref $value eq "ARRAY" )
		{
			foreach (@{$value})
			{
				$newdata .= "$key=$_\n";
			}
			next;
		}

		$newdata .= "$key=$value\n";
	}

	$newdata .= "\n";

	# read settings data
	if ( open F, "+<$CFG_APPSETUP_FILE" )
	{
		flock F, 2; # LOCK_EX
		$content = join ("", <F>);
		$content =~ s/\r//g;
		truncate F, 0;
		seek F, 0, 0;
	}
	else
	{
		open F, ">$CFG_APPSETUP_FILE" or return;
	}

	$content = $content =~ /\[$appid\]\n.*?\n\n/s
			? $` . $newdata . $'
			: "$content$newdata";

	print F $content;
	flock F, 8; # LOCK_UN
	close F;

	return 1;
}


# Removes application info for specified ID
sub remove_app_info
{
	my $appid = shift;

	# read settings data
	open F, "+<$CFG_APPSETUP_FILE" or return;
	flock F, 2; # LOCK_EX
	my $content = join ("", <F>);
	$content =~ s/\r//g;

	if ( $content =~ /\[$appid\]\n.*?\n\n/s )
	{
		$content = $` . $';
		# and store them
		truncate F, 0;
		seek F, 0, 0;
		print F $content;
	}

	flock F, 8; # LOCK_UN
	close F;

	unlink $CFG_APPSETUP_FILE if -s $CFG_APPSETUP_FILE == 0;
	return 1;
}

# returns directory component from the path
sub internal_dirname
{
	return ($_[0] =~ /(.*)\/[^\/]+$/) ? $1 : $_[0];
}

# Return machine information - arch, os name, os version
sub get_os_info
{
	my $arch = `uname -p 2>/dev/null`;
	my $rev = `uname -r`;
	my $name = `uname -s`;
	$arch = `uname -m` if !$arch || $arch =~ /^unknown/;

	$arch =~ s/[\s\r\n]//g;
	$rev =~ s/[\s\r\n]//g;
	$name =~ s/[\s\r\n]//g;

    # Special version handling for FreeBSD, OpenBSD...
	$rev = $1 if $name =~ /bsd/i && $rev =~ /^([\d\.]+)/;

	# hey, this is Linux!
	return ($arch, $name, $rev);
}


# Display a simple help about command-line options
sub display_cmdline_options
{
	my %optlist = @_;
	print "Supported command line options:\n";
	foreach (keys %optlist)
	{
		next if !$optlist{$_};
		print ("    $_ - $optlist{$_}\n");
	}
}

# checks for root permissions, and aborts if not
sub check_for_root
{
	my $procname = shift;

	return if $^O eq "cygwin" || $^O eq "MSWin32";
	klinstall::fatal ("%1 must be run by root!", $procname) if $> != 0;
}


# Does a file lookup in all the paths specified in argument list.
sub file_lookup
{
	my ($fn, @paths) = @_;

	foreach my $pathlist (@paths)
	{
		my $path = "$pathlist/$fn";
		$path =~ s/\/\//\//g;
		return $path if -f $path;
	}

	return;
}


# Get a pid(s) of a running process with specified name, or undef
sub get_process_pids
{
	my $processname = shift;
	my (%pids, @pidlist);

	# ps options and patterns are different on different systems
	my @osdata = get_os_info();

	if ( $osdata[1] =~ /hpux|sunos/i )
	{
		my @plist = split ("\n", `ps -ef`);
		shift @plist;
		foreach (@plist)
		{
			my $pid_part = substr ($_, 9, 6);
			my $cmd_part = substr ($_, 47);
			next if $pid_part !~ /\s*(\d+)\s$/;
			$pid_part = $1;

			$pids{$pid_part} = $cmd_part;
			$pids{$pid_part} = $1 if $pids{$pid_part} =~ /^(.*?)\s+/;
		}
	}
	else
	{
		my @plist = split ("\n", `ps ax`);
		shift @plist;
		foreach (@plist)
		{
			s/\s+$//;
			next if !/^\s*(\d+)/;
			my $pid_part = $1;
			my $cmd_part = substr ($_, 27);
			$pids{$pid_part} = $cmd_part;

			$pids{$pid_part} = $1 if $pids{$pid_part} =~ /^(.*?)\s+/;
		}
	}

	foreach (keys %pids)
	{
		next if $pids{$_} !~ /\b$processname\b/;
		push @pidlist, $_;
	}

	return @pidlist;
}


# Make recursive path. Mode argument is optional
sub make_path
{
	my ($dirname, $mode, $trackref) = @_;

	return if -d $dirname;
	if ( $dirname =~ /(.+)\/[^\/]+$/ )
	{
		my $msg = make_path ($1, $mode, $trackref);
		return $msg if $msg;
	}

	return "Could not create directory $dirname: $!"
		if !mkdir ($dirname, $mode);

	push (@{$trackref}, $dirname) if $trackref;
	return;
}


# Remove recursive path. Mode argument is optional
sub remove_path
{
	my $dirname = shift;

	return 1 if !-d $dirname;
	remove_path ($1) if $dirname =~ /\/([\/+])$/;

	warning "Could not delete directory $dirname: $!\n" if !rmdir $dirname;
}

# Remove recursive files tree.
sub rm_tree
{
	my ($path, $opt) = @_;

	return unlink $path if -f $path;

	return 0 if !-d $path || !opendir(DIR, $path);
	while(my $name = readdir(DIR)) {
		next if $name =~ /^\.\.?$/;

		if (-d "$path/$name") {
			return 0 if !rm_tree("$path/$name");
		} else {
			return 0 if !unlink("$path/$name");
		}
	}
	if ($opt) {
		return rmdir $path;
	}
	return 1;
}

# Copies a file. Returns the error message if failed
sub copy_file
{
	my ($src, $dest) = @_;

	# if $dest is a directory, add the source filename
	$dest .= "/$1" if -d $dest && $src =~ /([^\/]+)$/;

	open FSRC, "<$src" or return "Could not read file $src: $!";
	open FDST, ">$dest" or return "Could not write file $dest: $!";

	binmode FSRC;
	binmode FDST;
	my $length;

	while ( ($length = sysread (FSRC, my $data, 8192)) > 0 )
	{
		if ( syswrite (FDST, $data, $length) != $length )
		{
			close FSRC;
			close FDST;

			unlink $dest;
			return "Could not write file $dest: $!";
		}
	}

	close FSRC;
	close FDST;

	if ( !defined $length )
	{
		unlink $dest;
		return "Couldn't read file $src: $!";
	}

	# Set the same owner and mode
	my @statinfo = stat ($src);
	if ( @statinfo )
	{
		chown $statinfo[4], $statinfo[5], $dest;
		chmod (($statinfo[2] & 07777), $dest);
	}

	return;
}


# Copies the matched files. Returns the error messages array if failed
sub copy_files
{
	my ($srcdir, $destdir, $pattern) = @_;
	my @errormessages;

	if ( opendir (DIR, $srcdir) )
	{
		my @files = grep { -f "$srcdir/$_" && /$pattern/i } readdir DIR;
		closedir DIR;

		foreach (@files)
		{
			my $msg = klinstall::copy_file ("$srcdir/$_", "$destdir/$_");
			push @errormessages, $msg if $msg;
		}
	}
	else
	{
		push @errormessages, "Could not open directory $srcdir: $!";
	}

	return @errormessages;
}


# Alters the file/dir owner/group
sub lazy_chown
{
	my $usergroup = shift;
	my @targets = @_;

	my @arr = split (/:/, $usergroup);
	my $user = ($arr[0] && getpwnam($arr[0])) ? getpwnam($arr[0]) : -1;
	my $group = ($arr[1] && getgrnam($arr[1])) ? getgrnam($arr[1]) : -1;

	# Do we need to change ownership?
	return if $user == -1 && $group == -1;

	chown ($user, $group, @targets);
}


# Copies a file, and sets up its owner/permission
sub install_file
{
	my ($src, $dest, $mode, $owner) = @_;

	my $msg = copy_file ($src, $dest);
	fatal ($msg) if $msg;
	lazy_chown ($owner, $dest);
	chmod (oct($mode), $dest) if $mode;
}


# Creates a directory, and sets up its permission
sub install_path
{
	my ($dir, $mode, $owner, $trackref) = @_;

	my $msg = make_path ($dir, oct($mode), $trackref);
	fatal ($msg) if $msg;

	lazy_chown ($owner, $dir);
	chmod (oct($mode), $dir);
}


# Asks a question from user, or from autoanswer file
sub ask_question
{
	my ($qid, $def_answer, $question, @args) = @_;

	# don't ask anything if question is already answered
	return $AUTOANSWERS{$qid} if defined $AUTOANSWERS{$qid};

	my $qtext = translate_string ($question, @args);
	chomp ($qtext);

	output ($def_answer ? "$qtext [$def_answer]: " : $qtext);

	my $answer = <STDIN>;
	$answer = "" if !defined $answer;
	$answer =~ s/[\r\n]//g;
	$answer = $def_answer if length($answer) == 0;

	# For debug mode: record the question and the answer in the appropriate file
	if ( defined $ENV{"KAV_INSTALL_HISTORY_FILE"} && open F, ">>$ENV{KAV_INSTALL_HISTORY_FILE}" )
	{
		print F "$qid=$answer\n";
		close F;
	}

	output "\n";
	return $answer;
}


# Ask a 'yes/no' question
sub ask_boolean
{
	while ( 1 )
	{
		my $answ = ask_question (@_);

		$answ = "yes" if $answ =~ /^yes$/i || $answ =~ /^y/i;
		$answ = "no"  if $answ =~ /^no$/i  || $answ =~ /^n/i;

		return 1 if $answ eq "yes";
		return 0 if $answ eq "no";

		output "Please answer either 'yes' or 'no'.\n";
	}
}


# Parses command-line args and script name, gets the work-logic
sub get_work_logic
{
	my ($parsename, $argsref) = @_;
	my ($pkgname, $part) = ("", "");

	# cut the path
	$parsename = $1 if $parsename =~ /([^\/]+)$/;

	# if argv[0] is package manager name, get it
	if ( @{$argsref} && defined $PKGMGRS{@{$argsref}[0]} )
	{
		$pkgname = $PKGMGRS{@{$argsref}[0]};
		shift @{$argsref};
	}

	# if argv[1] is logic part name, get it
	if ( @{$argsref}
	&& (@{$argsref}[0] eq "pre" || @{$argsref}[0] eq "post") )
	{
		$part = @{$argsref}[0];
		shift @{$argsref};
		return ($part, $pkgname);
	}

	# parse name
	$part = "pre" if $parsename =~ /^pre/;
	$part = "post" if $parsename =~ /^post/;

	return ($part, $pkgname);
}


# Check the source directory for valid AV bases set, and install them
sub install_AV_bases
{
	my ($srcdir, $destdir) = @_;

	open ASET, "<$srcdir/avp.set" or return;
	my @set_content = <ASET>;
	close ASET;

	# check for all the files existance
	foreach (@set_content)
	{
		s/[\n\r]//g;
		next if /^[;#]/ || /^$/;
		return if !-f "$srcdir/$_";
	}

	# and copy the files
	klinstall::output ("Found Kaspersky Anti-Virus bases in $srcdir, copying them.");
	klinstall::copy_files ($srcdir, $destdir, ".+");

	return 1;
}


# execute a system command
sub exec_cmd
{
	my $cmd = shift;
	my $msg;

	# Escape shell args
	$cmd =~ s/([\&;\`'\\\|"*?~<>^\(\)\[\]\{\}\$\n\r])/\\$1/g;

	my $tmpfile = "/tmp/klinstallexec.$$.tmp";
	my $retcode = system ("$cmd 1>$tmpfile 2>&1");
	$retcode = ($retcode == -1 ? "undef" : $?); # -1 means system() failed

	if ( open FILE, "<$tmpfile" )
	{
		$msg = join ("", <FILE>);
		close FILE;

		unlink $tmpfile;
	}

	return ($retcode, $msg);
}


# execute a system command, escaping arguments
sub exec_cmd_safe
{
	my @args = @_;
	my ($msg, $cmd);

	# Escape shell args
	foreach (@args)
	{
		s/([\&;\`'\\\| "*?~<>^\(\)\[\]\{\}\$\n\r])/\\$1/g;
	}

	my $tmpfile = "/tmp/klinstallexec.$$.tmp";
	my $cmdline = join (" ", @args) . " 1>$tmpfile 2>&1";

	my $retcode = system ($cmdline);
	$retcode = ($retcode == -1 ? "undef" : $?); # -1 means system() failed

	if ( open FILE, "<$tmpfile" )
	{
		$msg = join ("", <FILE>);
		close FILE;

		unlink $tmpfile;
	}

	return ($retcode, $msg);
}


# execute a system command. If the command doesn't exit with code 0, abort
# with the error message
sub exec_zero
{
	my ($rc, $msg) = exec_cmd (@_);

	if ( !defined $rc		# system() failed
	|| ($? >> 8) != 0		# exit code != 0
	|| ($? & 127) != 0 )	# signal num
	{
		fatal ("execute ($_[0]) failed: $msg");
	}
}


# Ini file error message
sub get_ini_error
{
	return $INI_ERROR_TEXT ? $INI_ERROR_TEXT : undef;
}

# Load some data from the configuration (ini-like) file
sub load_ini_file
{
	if ( !open FILE, "<$_[0]" )
	{
		$INI_ERROR_TEXT = "unable to read file $_[0]: $!";
		return;
	}

	my $content = join ("", <FILE>);
	close FILE;

	# just in case
	$content =~ s/\r//g;

	my ($sname, $sdata) = ("&%HEADER%&", "");
	my (%inicontent);

	foreach my $line (split ("\n", $content))
	{
		# section header
		if ( $line =~ /^\s*\[(.*?)\]\s*/ )
		{
			if ( $sname ne "&%HEADER%&" )
			{
				$inicontent{"sections"}{$sname} = $sdata;
				push @{$inicontent{"order"}}, $sname;
			}
			else
			{
				$inicontent{"header"} = $sdata;
			}

			($sname, $sdata) = ($1, "");
			next;
		}

		$sdata .= $line . "\n";
	}

	if ( $sname ne "&%HEADER%&" )
	{
		$inicontent{"sections"}{$sname} = $sdata;
		push @{$inicontent{"order"}}, $sname;
	}
	else
	{
		$inicontent{"header"} = $sdata;
	}

	# no sections
	if ( !$inicontent{"order"} || !@{$inicontent{"order"}} )
	{
		$INI_ERROR_TEXT = "no sections found in file $_[0]";
		return;
	}

	return %inicontent;
}


sub save_ini_file
{
	my $filename = shift;
	my %inidata = @_;

	if ( !open FILE, ">$filename" )
	{
		$INI_ERROR_TEXT = "unable to write file $filename: $!";
		return;
	}

	print FILE "$inidata{header}" if $inidata{"header"};

	foreach my $name (@{$inidata{"order"}})
	{
    	print FILE "[$name]\n$inidata{sections}->{$name}";
    	print FILE "\n" if $inidata{"sections"}->{$name} !~ /\n\n$/;
	}

    close FILE;
	return 1;
}


# Get ini file value
sub get_ini_value
{
	my ($ref, $section, $key) = @_;

	my $undefpattern = "^$key\\s*=\\s*\$";
	my $pattern = "^$key\\s*=\\s*(.+?)\$";

	return undef if !$ref->{"sections"}->{$section}
		|| $ref->{"sections"}->{$section} =~ /$undefpattern/m
		|| $ref->{"sections"}->{$section} !~ /$pattern/m;

	return $1;
}


# Set ini file value
sub set_ini_value
{
	my ($ref, $section, %data) = @_;

	# add the section if not exist
	if ( !$ref->{"sections"}->{$section} )
	{
		push @{$ref->{"order"}}, $section;
		$ref->{"sections"}->{$section} = "";
	}

	# search through the content
	foreach my $key (keys %data)
	{
		my $value = defined $data{$key} ? $data{$key} : "";

		$ref->{"sections"}->{$section} =~ m/^\s*$key\s*=\s*/m
			and $ref->{"sections"}->{$section} =~ s/^\s*$key\s*=.*?\n/$key=$value\n/m
				or $ref->{"sections"}->{$section} = "$key=$value\n$ref->{sections}->{$section}";
	}
}


# Delete the whole section, or ini file key=value
sub remove_ini_value
{
	my ($ref, $section, @keys) = @_;

	return if !$ref->{"sections"}->{$section};

	foreach my $key (@keys)
	{
		$ref->{"sections"}->{$section} =~ s/^\s*$key\s*=\s*.*\n?//gm;
	}

	# add delete the section if empty
	if ( !@keys || $ref->{"sections"}->{$section} =~ /^\s*$/s )
	{
		delete $ref->{"sections"}->{$section};
		@{$ref->{"order"}} = grep { $_ ne $section } @{$ref->{"order"}};
	}
}


# Modify some data in the configuration (ini-like) file
sub update_ini_file
{
	my ($filename, $section, %data) = @_;

	# read inifile
	my %inidata = load_ini_file ($filename);
	return if !%inidata;

	set_ini_value (\%inidata, $section, %data);

	return save_ini_file ($filename, %inidata);
}


# Creates a new user account if not exist, returns undef if OK, or the error message text.
# It also creates homedir if it is not exist.
sub add_user
{
	my ($username, $group, $shell, $home) = @_;

	return 1 if getpwnam ($username);
	my @cmd = $IS_FREEBSD ? ("pw", "useradd", "-n", $username) : ("useradd");

	push @cmd, ("-g", $group) if $group;
	push @cmd, ("-G", $group) if $group and !$IS_FREEBSD;
	push @cmd, ("-s", $shell) if $shell;
	push @cmd, ("-d", $home) if $home;

	push @cmd, $username if !$IS_FREEBSD;

	make_path ($home, 0755) if !-d $home;

	return 1 if system (@cmd) == 0;
	$@ = "Could not add user '$username'";
	return undef;
}

sub add_group
{
	my $group = shift;

	return 1 if getgrnam ($group);
	my @cmd = $IS_FREEBSD ? ("pw", "groupadd", "-n", $group) : ("groupadd", $group);

	return 1 if system (@cmd) == 0;
	$@ = "Could not add group '$group'";
	return undef;
}


sub add_user_to_group
{
	my ($user, $group) = @_;

	# Check whether the user is already in this secondary group
	my $groupmembers = ((getgrnam ($user))[3]);
	my $userrx = quotemeta ($user);
	my @cmd;

	return 1 if $groupmembers && $groupmembers =~ /\b$userrx\b/;

	if ( $IS_FREEBSD ) {
		@cmd = ("pw", "usermod", $user, "-G", $group);
	} elsif ( $^O =~ /^openbsd$/i ) {
		@cmd =("usermod", "-G", $group, $user);
	} else {
		@cmd = ("groupmod", "-A", $user, $group);
	}

	return 1 if system (@cmd) == 0;
	$@ = "Could not add user '$user' to group 'group'";
	return undef;
}

sub exec_cmd_as_user
{
	my ($user, @args) = @_;
	my $group;

	if ($user =~ /^(\w+):(\w+)$/) {
	    $user = $1;
	    $group = $2;
	}

	my $uid = getpwnam($user);

	return 254 if !defined $uid;

	my $gid;
	if ( defined($group) ) {
	    $gid = getgrnam($group);
	    return 254 if !defined $gid;
	}

	my $pid = fork();

	return -1 if $pid == -1;

	if ( !$pid )
	{
		# child
		if (defined($gid)) {
		    $! = 0;
		    $( = $gid;
	    	    $) = $gid;

		    exit 253 if $! != 0;
		}

		$! = 0;
		$< = $uid;
		$> = $uid;

		exit 253 if $! != 0;

		# reopen file descriptors
		open (STDOUT, "> /dev/null");
		open (STDERR, "> /dev/null");

		exec @args;
		exit 252;
	}

	wait();
	return $?;
}

# make rules for config conversion
sub load_rules_file {
	my $file = shift;
	my %rules=();

	open(RULES,"<$file") || fatal("can't open $file for read: $!");
	while(<RULES>) {
		s/[\r\n]+$//;

		next if /^#/ || /^\s*$/;

		fatal "wrong line in $file:$.:$_" if !/^(\S+)\s+(\S+)\s+$RULES_SPLIT_REPLACE_STR\s+(\S+)\s+(\S+)\s+$RULES_SPLIT_FROM_STR\s+\/(.*)\/\s+$RULES_SPLIT_TO_STR\s+\/(.*)\/(\s+([A-Z]+))?\s*$/;

		my ($_1, $_2, $_3, $_4, $_5, $_6, $_8, $is_regex) = ($1, $2, $3, $4, $5, $6, $8, 0);

		($_1, $is_regex) = (($_1 =~ /^\/(.*)\/$/) ? ($1, 1) : ($_1, 0));

		# search section-var index
		my $r_idx = ((exists($rules{$_3}) && exists($rules{$_3}{$_4})) ? scalar(@{$rules{$_3}{$_4}}) : 0) - 1;
		for(;$r_idx >= 0; --$r_idx){
			my $r = 0;
			eval { $r = (($rules{$_3}{$_4}->[$r_idx]{from_section} eq $_1) && ($rules{$_3}{$_4}->[$r_idx]{from_var} eq $_2)); };
			if ($@) { print "$_3->$_4\n".Dumper($rules{$_3}{$_4})."\n";die $@;}
			last if $r;
		}

		if ($r_idx < 0) {
			# add new section-var rule
			my %rule_hash = (
				from_section => $_1,
				is_regex => $is_regex,
				from_var => $_2
			);
			unshift(@{$rule_hash{rule}}, {from => $_5, to => $_6, type => $_8});
			unshift(@{$rules{$_3}{$_4}}, {%rule_hash});

		} else {
			# add rule into existing section-var
			unshift(@{$rules{$_3}{$_4}->[$r_idx]{rule}}, {from => $_5, to => $_6, type => $_8});
		}
	}
	close(RULES);
	return %rules;
}

# create new group config using old values
sub convert_group_config {
	my ($ref_ini, $ref_default_ini, $ref_rule, $default_group_name, $group_name, $in_group_file, $ref_const_value) = @_;

	my $OUTPUT = "";
	my$tmp_OUTPUT = "";

	my ($s_name, $v_name, $v_value);
	open(IN_CONF, "<$in_group_file") ||
			fatal("can't open group file $in_group_file for read: $!");

	while(<IN_CONF>) {
		s/[\r\n]+$//;

		if (/^#/ || /^\s*$/) {

			$tmp_OUTPUT = "$_\n";
		} else {

			if (/^\[(.*)\]/) {

				# have section name
				$s_name = $1;
				my $new_s_name = $s_name;
				$new_s_name =~ s/(^|\.|:)$default_group_name($|\.|:)/$1$group_name$2/i;

				$tmp_OUTPUT = "[$new_s_name]\n";

			} elsif (/^(\w+)\s*=\s*(.*)\s*$/) {

				# have var = value
				($v_name, $v_value) = ($1, $2);
				$tmp_OUTPUT = "$v_name=$v_value\n";

				if (defined($ref_const_value) && exists($ref_const_value->{$v_name})) {
					$tmp_OUTPUT = "$v_name=$ref_const_value->{$v_name}\n";

				} elsif (exists($ref_rule->{$s_name}) && exists($ref_rule->{$s_name}{$v_name})) {

					# search for section and var convert from
					for my $rule_hash (@{$ref_rule->{$s_name}{$v_name}}) {

						if ($rule_hash->{is_regex}) {
							my @values = ();

							# get source values
							for my $s_name_ini (keys %$ref_ini) {
								if ($s_name_ini =~ /$rule_hash->{from_section}/i) {
									my $ref_src_value = get_value_ignore_case($ref_ini->{$s_name_ini}, $rule_hash->{from_var});
									@values = @$ref_src_value if $ref_src_value;
									last;
								}
							}

							# try search next rule cose no values in source ini found
							next if !@values;

							# if we find at least one value, it must be equivalent to at least one rule.
							my $err = 0;
							$tmp_OUTPUT = "";
							for my $val (@values) {
								# for each value check for _FROM_ rule
								my $r_idx = @{$rule_hash->{rule}};
								while(--$r_idx >= 0) { last if $val =~ /$rule_hash->{rule}->[$r_idx]{from}/; }
								if ($r_idx < 0) {
									# need some warnings?
									klinstall::warning("<B>cannot be converted:</B>'\n[$rule_hash->{from_section}] $rule_hash->{from_var}=$val'\n");
									$err = 1;
									$tmp_OUTPUT .= "$v_name=$v_value\n";
								} else {
									$tmp_OUTPUT .= "$v_name=".join("\n$v_name=",convert_val($val, $rule_hash->{rule}->[$r_idx]))."\n";
								}
							}

							# stop search rule. no one rule found. maybe wrong value in source ini?
							last if $err;

						} else {
							my $ref_src_section = get_value_ignore_case($ref_ini, $rule_hash->{from_section});

							if ($ref_src_section) {
								my $ref_src_value = get_value_ignore_case($ref_src_section, $rule_hash->{from_var});

								if ($ref_src_value) {
									# if we find at least one value, it must be equivalent to at least one rule.
									my $err = 0;
									$tmp_OUTPUT = "";
									for my $val_ (@$ref_src_value) {
										# for each value check for _FROM_ rule
										my $r_idx = @{$rule_hash->{rule}};
										while(--$r_idx >= 0) { last if $val_ =~ /$rule_hash->{rule}->[$r_idx]{from}/; }
										if ($r_idx < 0) {
											$tmp_OUTPUT .= "$v_name=$v_value\n";
											if(length $val_) {
												# need some warnings?
												klinstall::warning("<B>cannot be converted:</B>'\n[$rule_hash->{from_section}] $rule_hash->{from_var}=$val_'\n");
												$err = 1;
											}
										} else {
											$tmp_OUTPUT .= "$v_name=".join("\n$v_name=",convert_val($val_, $rule_hash->{rule}->[$r_idx]))."\n";
										}
									}

									# stop search rule. no one rule found. maybe wrong value in source ini?
									last if $err;
								}
							}
						}
					}
				}
			}
		}
		$OUTPUT .= $tmp_OUTPUT;
	}
	close(IN_CONF);

	return $OUTPUT;
}

# convert main config without any includes
sub convert_main_config {
	my ($ref_ini, $ref_rule, $in_config_file, $out_config_file) = @_;

	my ($s_name, $v_name, $v_value);
	my $OUTPUT = "";
	my $tmp_OUTPUT = "";

	open(IN_CONF, "<$in_config_file") ||
			fatal("can't open $in_config_file for read: $!");

	while(<IN_CONF>) {
		s/[\r\n]+$//;

		if (/^#/ || /^\s*$/) {
			$tmp_OUTPUT = "$_\n";

		} else {

			if (/^\[(.*)\]/) {

				# have section name
				$s_name = $1;
				$tmp_OUTPUT = "$_\n";

			} elsif (/^(\w+)\s*=\s*(.*)\s*$/) {

				# have var = value
				($v_name, $v_value) = ($1, $2);
				$tmp_OUTPUT = "$v_name=$v_value\n";

				if (exists($ref_rule->{$s_name}) && exists($ref_rule->{$s_name}{$v_name})) {

					# search for section and var convert from
					for my $rule_hash (@{$ref_rule->{$s_name}{$v_name}}) {
						my $ref_src_section = get_value_ignore_case($ref_ini, $rule_hash->{from_section});

						if ($ref_src_section) {
							my $ref_src_value = get_value_ignore_case($ref_src_section, $rule_hash->{from_var});

							if ($ref_src_value) {
								# if we find at least one value, it must be equivalent to at least one rule.
								my $err = 0;
								$tmp_OUTPUT = "";
								for my $val_ (@$ref_src_value) {
									# for each value check for _FROM_ rule
									my $r_idx = @{$rule_hash->{rule}};
									while(--$r_idx >= 0) { last if $val_ =~ /$rule_hash->{rule}->[$r_idx]{from}/; }
									if ($r_idx < 0) {
									# need some warnings?
										klinstall::warning("<B>cannot be converted:</B>'\n[$rule_hash->{from_section}] $rule_hash->{from_var}=$val_'\n");
										$err = 1;
										$tmp_OUTPUT .= "$v_name=$v_value\n";
									} else {
										$tmp_OUTPUT .= "$v_name=".join("\n$v_name=",convert_val($val_, $rule_hash->{rule}->[$r_idx]))."\n";
									}
								}
								# stop search rule. no one rule found. maybe wrong value in source ini?
								last if $err;
							}
						}
					}
				}
			} else {
				fatal "wrong line in $in_config_file:$.:$_";
			}
		}
		$OUTPUT .= $tmp_OUTPUT;
	}
	close(IN_CONF);

	return $OUTPUT;
}

# get hash value by key ignoring key case
sub get_value_ignore_case {
	my ($ref_hash, $key) = @_;

	$key = (grep { /^$key$/i } keys %$ref_hash)[0];
	return $ref_hash->{$key} if $key;
}

# convert value by perl regex
sub convert_val {
	my ($v_val, $ref_rule_val) = @_;
	my @ret = ();

	if ($v_val =~ /$ref_rule_val->{from}/i) {
		my @rv = ();
		if (defined($ref_rule_val->{type})) {
			if ($ref_rule_val->{type} eq $RULES_MULT_STR) {
				while (@rv = $v_val =~ /$ref_rule_val->{from}/i) {
					$v_val = $';
					my $to_val = $ref_rule_val->{to};
					my $i = 1;
					while ($i <= @rv) {
						$to_val =~ s/\$$i/$rv[$i - 1]/ig;
						++$i;
					}
					push(@ret, $to_val);
				}
				return @ret;
			} elsif ($ref_rule_val->{type} eq $RULES_MULTLINE_STR) {
				while (@rv = $v_val =~ /$ref_rule_val->{from}/i) {
					$v_val = $';
					my $to_val = $ref_rule_val->{to};
					my $i = 1;
					while ($i <= @rv) {
						$to_val =~ s/\$$i/$rv[$i - 1]/ig;
						++$i;
					}
					push(@ret, $to_val);
				}
				return join(", ",@ret);
			}
		} else {
			@rv = $v_val =~ /$ref_rule_val->{from}/i;
			my $i = 1;
			my $tmp_v = $ref_rule_val->{to};

			while($i <= @rv) {
				$tmp_v =~ s/\$$i/$rv[$i - 1]/ig;
				++$i;
			}
			return $tmp_v;
		}
	}

	return $v_val;
}
